/*-
 * Copyright (c) 2001 Michael Smith
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$FreeBSD$
 */

/*
 * Common Interface for SCSI-3 Support driver.
 *
 * CISS claims to provide a common interface between a generic SCSI
 * transport and an intelligent host adapter.
 *
 * This driver supports CISS as defined in the document "CISS
 * Command Interface for SCSI-3 Support Open Specification", Version
 * 1.04, Valence Number 1, dated 20001127, produced by Compaq Computer
 * Corporation.  This document appears to be a hastily and somewhat
 * arbitrarlily cut-down version of a larger (and probably even more
 * chaotic and inconsistent) Compaq internal document.  Various details
 * were also gleaned from Compaq's "cciss" driver for Linux.
 *
 * We provide a shim layer between the CISS interface and CAM,
 * offloading most of the queueing and being-a-disk chores on CAM.
 * Entry to the driver is via the PCI bus attachment (ciss_probe,
 * ciss_attach, etc) and via the CAM interface (ciss_cam_action,
 * ciss_cam_poll).
 *
 * Note that we currently only support the "simple" transport layer
 * over PCI.  This interface (ab)uses the I2O register set (specifically
 * the post queues) to exchange commands with the adapter.
 * 
 * Currently the only support CISS adapters are the Compaq Smart Array
 * 5* series (5300, 5i, 532).  Even with only three adapters, Compaq
 * still manage to have interface variations. 8(
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/devicestat.h>
#include <sys/stat.h>

#include <cam/cam.h>
#include <cam/cam_ccb.h>
#include <cam/cam_periph.h>
#include <cam/cam_sim.h>
#include <cam/cam_xpt_sim.h>
#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_message.h>

#include <machine/clock.h>
#include <machine/bus_memio.h>
#include <machine/bus.h>
#include <machine/endian.h>
#include <machine/resource.h>
#include <sys/rman.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#include <dev/ciss/cissreg.h>
#include <dev/ciss/cissvar.h>

MALLOC_DEFINE(CISS_MALLOC_CLASS, "ciss_data", "ciss internal data buffers");

/* pci interface */
static int	ciss_lookup(device_t dev);
static int	ciss_probe(device_t dev);
static int	ciss_attach(device_t dev);
static int	ciss_detach(device_t dev);
static int	ciss_shutdown(device_t dev);

/* (de)initialisation functions */
static int	ciss_init_pci(struct ciss_softc *sc);
static int	ciss_wait_adapter(struct ciss_softc *sc);
static int	ciss_init_requests(struct ciss_softc *sc);
static void	ciss_command_map_helper(void *arg, bus_dma_segment_t *segs, int nseg, int error);
static int	ciss_init_logical(struct ciss_softc *sc);
static void	ciss_free(struct ciss_softc *sc);

/* request submission/completion */
static int	ciss_start(struct ciss_request *cr);
static void	ciss_done(struct ciss_softc *sc);
static void	ciss_intr(void *arg);
static void	ciss_complete(struct ciss_softc *sc);
static int	ciss_report_request(struct ciss_request *cr, int *command_status, int *scsi_status);
static int	ciss_poll_request(struct ciss_request *cr, int timeout);
static int	ciss_wait_request(struct ciss_request *cr, int timeout);

/* request queueing */
static int	ciss_get_request(struct ciss_softc *sc, struct ciss_request **crp);
static void 	ciss_release_request(struct ciss_request *cr);

/* DMA map/unmap */
static int	ciss_map_request(struct ciss_request *cr);
static void	ciss_request_map_helper(void *arg, bus_dma_segment_t *segs, int nseg, int error);
static void	ciss_unmap_request(struct ciss_request *cr);

/* CAM interface */
static int	ciss_cam_init(struct ciss_softc *sc);
static void	ciss_cam_rescan(struct ciss_softc *sc);
static void	ciss_cam_rescan_callback(struct cam_periph *periph, union ccb *ccb);
static void	ciss_cam_action(struct cam_sim *sim, union ccb *ccb);
static int	ciss_cam_action_io(struct cam_sim *sim, struct ccb_scsiio *csio);
static void	ciss_cam_backoff_ccb(struct ciss_softc *sc, struct ccb_scsiio *csio);
static void	ciss_cam_poll(struct cam_sim *sim);
static void	ciss_cam_complete(struct ciss_request *cr);

/* periodic status monitoring */
static void	ciss_periodic(void *arg);
static void	ciss_notify_event(struct ciss_softc *sc, struct ciss_request *cr);
static void	ciss_notify_complete(struct ciss_request *cr);

/* debugging output */
static void	ciss_print_adapter(struct ciss_softc *sc);
static void	ciss_print_request(struct ciss_request *cr);

/*
 * PCI bus interface.
 */
static device_method_t ciss_methods[] = {
    /* Device interface */
    DEVMETHOD(device_probe,	ciss_probe),
    DEVMETHOD(device_attach,	ciss_attach),
    DEVMETHOD(device_detach,	ciss_detach),
    DEVMETHOD(device_shutdown,	ciss_shutdown),
    { 0, 0 }
};

static driver_t ciss_pci_driver = {
    "ciss",
    ciss_methods,
    sizeof(struct ciss_softc)
};

static devclass_t	ciss_devclass;
DRIVER_MODULE(ciss, pci, ciss_pci_driver, ciss_devclass, 0, 0);

/*
 * Control device interface.
 */
static d_open_t		ciss_open;
static d_close_t	ciss_close;
static d_ioctl_t	ciss_ioctl;

#define CISS_CDEV_MAJOR  158

static struct cdevsw ciss_cdevsw = {
    ciss_open, ciss_close, noread, nowrite, ciss_ioctl,
    nopoll, nommap, nostrategy, "ciss", CISS_CDEV_MAJOR,
    nodump, nopsize, 0, -1
};

/************************************************************************
 * CISS adapters amazingly don't have a defined programming interface value.
 * (One could say some very despairing things about PCI and people just
 * not getting the general idea.)  So we are forced to stick with matching
 * against subvendor/subdevice.
 */
#define CISS_BOARD_SA5	(1<<0)
#define CISS_BOARD_SA5B	(1<<1)

static struct
{
    u_int16_t	subvendor;
    u_int16_t	subdevice;
    int		flags;
    char	*desc;
} ciss_vendor_data[] = {
    { 0x0e11, 0x4070, CISS_BOARD_SA5,	"Compaq Smart Array 5300" },
    { 0x0e11, 0x4080, CISS_BOARD_SA5B,	"Compaq Smart Array 5i" },
    { 0x0e11, 0x4082, CISS_BOARD_SA5B,	"Compaq Smart Array 532" },
    { 0, 0, NULL }
};

/************************************************************************
 * Find a match for the device in our list of known adapters.
 */
static int
ciss_lookup(device_t dev)
{
    int 	i;
    
    for (i = 0; ciss_vendor_data[i].desc != NULL; i++)
	if ((pci_get_subvendor(dev) == ciss_vendor_data[i].subvendor) &&
	    (pci_get_subdevice(dev) == ciss_vendor_data[i].subdevice)) {
	    return(i);
	}
    return(-1);
}

/************************************************************************
 * Match a known CISS adapter.
 */
static int
ciss_probe(device_t dev)
{
    int		i;
    
    i = ciss_lookup(dev);
    if (i != -1) {
	device_set_desc(dev, ciss_vendor_data[i].desc);
	return(-10);
    }
    return(ENOENT);
}	

/************************************************************************
 * Attach the driver to this adapter.
 */
static int
ciss_attach(device_t dev)
{
    struct ciss_softc	*sc;
    int			i, error;

    debug_called(2);

#ifdef CISS_DEBUG
    debug_struct(ciss_command);
    debug_struct(ciss_header);
    debug_union(ciss_device_address);
    debug_struct(ciss_cdb);
    debug_struct(ciss_report_cdb);
    debug_struct(ciss_notify_cdb);
    debug_struct(ciss_notify);
    debug_struct(ciss_message_cdb);
    debug_struct(ciss_error_info_pointer);
    debug_struct(ciss_error_info);
    debug_struct(ciss_sg_entry);
    debug_struct(ciss_config_table);
    debug_const(CISS_MAX_REQUESTS);
    debug_const(CISS_MAX_LOGICAL);
    debug_const(CISS_INTERRUPT_COALESCE_DELAY);
    debug_const(CISS_INTERRUPT_COALESCE_COUNT);
    debug_const(CISS_COMMAND_ALLOC_SIZE);
    debug_const(CISS_COMMAND_SG_LENGTH);
#endif

    sc = device_get_softc(dev);
    sc->ciss_dev = dev;

    /*
     * Work out controller type.
     */
    i = ciss_lookup(dev);
    if (ciss_vendor_data[i].flags & CISS_BOARD_SA5)
	sc->ciss_interrupt_mask = CISS_TL_SIMPLE_INTR_OPQ_SA5;
    if (ciss_vendor_data[i].flags & CISS_BOARD_SA5B)
	sc->ciss_interrupt_mask = CISS_TL_SIMPLE_INTR_OPQ_SA5B;
	
    /*
     * Do PCI-specific init.
     */
    if ((error = ciss_init_pci(sc)) != 0)
	goto out;
    /*
     * Initialise controller queues.
     */
    ciss_initq_free(sc);
    ciss_initq_busy(sc);
    ciss_initq_complete(sc);

    /*
     * Initialise command/request pool.
     */
    if ((error = ciss_init_requests(sc)) != 0)
	goto out;
    
    /*
     * Build our private table of logical devices.
     */
    if ((error = ciss_init_logical(sc)) != 0)
	goto out;

    /*
     * Initialise the CAM interface.
     */
    if ((error = ciss_cam_init(sc)) != 0)
	goto out;

    /*
     * Create the control device.
     */
    sc->ciss_dev_t = make_dev(&ciss_cdevsw, device_get_unit(sc->ciss_dev),
			      UID_ROOT, GID_OPERATOR, S_IRUSR | S_IWUSR,
			      "ciss%d", device_get_unit(sc->ciss_dev));
    sc->ciss_dev_t->si_drv1 = sc;

    /*
     * Enable interrupts so that the CAM scan can complete.
     */
    CISS_TL_SIMPLE_ENABLE_INTERRUPTS(sc);

    error = 0;
 out:
    if (error != 0)
	ciss_free(sc);
    return(error);
}

/************************************************************************
 * Detach the driver from this adapter.
 */
static int
ciss_detach(device_t dev)
{
    struct ciss_softc	*sc = device_get_softc(dev);
    
    /* XXX abort Notify Event command */

    /* XXX quiesce adapter */

    ciss_free(sc);
    return(0);
    
}

/************************************************************************
 * Prepare adapter for system shutdown.
 */
static int
ciss_shutdown(device_t dev)
{
    /* XXX quiesce adapter */
    return(0);
}

/************************************************************************
 * Perform PCI-specific attachment actions.
 */
static int
ciss_init_pci(struct ciss_softc *sc)
{
    uintptr_t		cbase, csize, cofs;
    int			error, i;

    debug_called(2);

    /*
     * Allocate register window first (we need this to find the config struct).
     */
    error = ENXIO;
    sc->ciss_regs_rid = CISS_TL_SIMPLE_BAR_REGS;
    if ((sc->ciss_regs_resource =
	 bus_alloc_resource(sc->ciss_dev, SYS_RES_MEMORY, &sc->ciss_regs_rid, 0, ~0, 1, RF_ACTIVE)) == NULL) {
	ciss_printf(sc, "can't allocate register window\n");
	return(ENXIO);
    }
    sc->ciss_regs_bhandle = rman_get_bushandle(sc->ciss_regs_resource);
    sc->ciss_regs_btag = rman_get_bustag(sc->ciss_regs_resource);
    
    /*
     * Find the BAR holding the config structure.  If it's not the one we already mapped for
     * registers, map it too.
     */
    sc->ciss_cfg_rid = CISS_TL_SIMPLE_READ(sc, CISS_TL_SIMPLE_CFG_BAR) & 0xffff;
    if (sc->ciss_cfg_rid != sc->ciss_regs_rid) {
	if ((sc->ciss_cfg_resource =
	     bus_alloc_resource(sc->ciss_dev, SYS_RES_MEMORY, &sc->ciss_cfg_rid, 0, ~0, 1, RF_ACTIVE)) == NULL) {
	    ciss_printf(sc, "can't allocate config window\n");
	    return(ENXIO);
	}
	cbase = (uintptr_t)rman_get_virtual(sc->ciss_cfg_resource);
	csize = rman_get_end(sc->ciss_cfg_resource) - rman_get_start(sc->ciss_cfg_resource) + 1;
    } else {
	cbase = (uintptr_t)rman_get_virtual(sc->ciss_regs_resource);
	csize = rman_get_end(sc->ciss_regs_resource) - rman_get_start(sc->ciss_regs_resource) + 1;
    }
    cofs = CISS_TL_SIMPLE_READ(sc, CISS_TL_SIMPLE_CFG_OFF);
    
    /*
     * Use the base/size/offset values we just calculated to sanity-check the
     * config structure.  If it's OK, point to it.
     */
    if ((cofs + sizeof(struct ciss_config_table)) > csize) {
	ciss_printf(sc, "config table outside window\n");
	return(ENXIO);
    }
    sc->ciss_cfg = (struct ciss_config_table *)(cbase + cofs);
    debug(3, "config struct at %p", sc->ciss_cfg);
    
    /*
     * Validate the config structure.
     * If we supported other transport methods, we could select amongst them
     * at this point in time.
     */
    if (strncmp(sc->ciss_cfg->signature, "CISS", 4)) {
	ciss_printf(sc, "config signature mismatch (got '%c%c%c%c')\n",
		    sc->ciss_cfg->signature[0], sc->ciss_cfg->signature[1],
		    sc->ciss_cfg->signature[2], sc->ciss_cfg->signature[3]);
	return(ENXIO);
    }
    if ((sc->ciss_cfg->valence < CISS_MIN_VALENCE) || (sc->ciss_cfg->valence > CISS_MAX_VALENCE)) {
	ciss_printf(sc, "adapter interface specification (%d) unsupported\n", 
		    sc->ciss_cfg->valence);
	return(ENXIO);
    }
    /*
     * Save the initial heartbeat value.
     */
    sc->ciss_heartbeat = sc->ciss_cfg->heartbeat;
    sc->ciss_heart_attack = 0;
    debug(3, "initial heartbeat 0x%x", sc->ciss_heartbeat);

    /*
     * Put the board into simple mode, and tell it we're using the low
     * 4GB of RAM.  Set the default interrupt coalescing options.
     */
    if (!(sc->ciss_cfg->supported_methods & CISS_TRANSPORT_METHOD_SIMPLE)) {
	ciss_printf(sc, "controller does not support 'simple' transport layer\n");
	return(ENXIO);
    }
    sc->ciss_cfg->requested_method = CISS_TRANSPORT_METHOD_SIMPLE;
    sc->ciss_cfg->command_physlimit = 0;
    sc->ciss_cfg->interrupt_coalesce_delay = CISS_INTERRUPT_COALESCE_DELAY;
    sc->ciss_cfg->interrupt_coalesce_count = CISS_INTERRUPT_COALESCE_COUNT;
    CISS_TL_SIMPLE_WRITE(sc, CISS_TL_SIMPLE_IDBR, CISS_TL_SIMPLE_IDBR_CFG_TABLE);
    error = ENXIO;
    for (i = 0; i < 1000; i++) {
	if (!(CISS_TL_SIMPLE_READ(sc, CISS_TL_SIMPLE_IDBR) & CISS_TL_SIMPLE_IDBR_CFG_TABLE)) {
	    error = 0;
	    break;
	}
	DELAY(1000);
    }
    if (error != 0) {
	ciss_printf(sc, "controller refuses to accept config update (IDBR 0x%x)\n",
		    CISS_TL_SIMPLE_READ(sc, CISS_TL_SIMPLE_IDBR));
	return(error);
    }
    if (!(sc->ciss_cfg->active_method != CISS_TRANSPORT_METHOD_SIMPLE)) {
	ciss_printf(sc, "controller refuses to go into 'simple' transport mode (0x%x, 0x%x)\n",
		    sc->ciss_cfg->supported_methods, sc->ciss_cfg->active_method);
	return(ENXIO);
    }
    if (1/*bootverbose*/)
	ciss_print_adapter(sc);

    /*
     * Turn off interrupts before we go routing anything.
     */
    CISS_TL_SIMPLE_DISABLE_INTERRUPTS(sc);
    
    /*
     * Allocate and set up our interrupt.
     */
    sc->ciss_irq_rid = 0;
    if ((sc->ciss_irq_resource =
	 bus_alloc_resource(sc->ciss_dev, SYS_RES_IRQ, &sc->ciss_irq_rid, 0, ~0, 1, 
			    RF_ACTIVE | RF_SHAREABLE)) == NULL) {
	ciss_printf(sc, "can't allocate interrupt\n");
	return(ENXIO);
    }
    if (bus_setup_intr(sc->ciss_dev, sc->ciss_irq_resource, INTR_TYPE_CAM, ciss_intr, sc,
		       &sc->ciss_intr)) {
	ciss_printf(sc, "can't set up interrupt\n");
	return(ENXIO);
    }

    /*
     * Allocate the parent bus DMA tag appropriate for our PCI interface.
     * 
     * Note that "simple" controllers can only address within a 32-bit span.
     */
    if (bus_dma_tag_create(NULL, 			/* parent */
			   1, 0, 			/* alignment, boundary */
			   BUS_SPACE_MAXADDR_32BIT,	/* lowaddr */
			   BUS_SPACE_MAXADDR, 		/* highaddr */
			   NULL, NULL, 			/* filter, filterarg */
			   MAXBSIZE, CISS_COMMAND_SG_LENGTH,	/* maxsize, nsegments */
			   BUS_SPACE_MAXSIZE_32BIT,	/* maxsegsize */
			   BUS_DMA_ALLOCNOW,		/* flags */
			   &sc->ciss_parent_dmat)) {
	ciss_printf(sc, "can't allocate parent DMA tag\n");
	return(ENOMEM);
    }

    /*
     * Create DMA tag for mapping buffers into controller-addressable space.
     */
    if (bus_dma_tag_create(sc->ciss_parent_dmat, 	/* parent */
			   1, 0, 			/* alignment, boundary */
			   BUS_SPACE_MAXADDR,		/* lowaddr */
			   BUS_SPACE_MAXADDR, 		/* highaddr */
			   NULL, NULL, 			/* filter, filterarg */
			   MAXBSIZE, CISS_COMMAND_SG_LENGTH,	/* maxsize, nsegments */
			   BUS_SPACE_MAXSIZE_32BIT,	/* maxsegsize */
			   0,				/* flags */
			   &sc->ciss_buffer_dmat)) {
	ciss_printf(sc, "can't allocate buffer DMA tag\n");
	return(ENOMEM);
    }
    return(0);
}

/************************************************************************
 * Wait for the adapter to come ready.
 */
static int
ciss_wait_adapter(struct ciss_softc *sc)
{
    int		i;

    debug_called(3);
    
    /*
     * Wait for the adapter to come ready.
     */
    if (!(sc->ciss_cfg->active_method & CISS_TRANSPORT_METHOD_READY)) {
	ciss_printf(sc, "waiting for controller to come ready...\n");
	for (i = 0; !(sc->ciss_cfg->active_method & CISS_TRANSPORT_METHOD_READY); i++) {
	    DELAY(1000000);	/* one second */
	    if (i > 30) {
		ciss_printf(sc, "timed out waiting for controller to come ready\n");
		return(EIO);
	    }
	}
    }
    return(0);
}

/************************************************************************
 * Allocate memory for the controller command structures, initialise
 * the request structures.
 *
 * Note that the entire set of commands are allocated in a single contiguous
 * slab.
 */
static int
ciss_init_requests(struct ciss_softc *sc)
{
    struct ciss_request	*cr;
    int			i;

    debug_called(2);
    
    /*
     * Calculate the number of request structures/commands we are going to
     * provide for this adapter.
     */
    sc->ciss_max_requests = min(CISS_MAX_REQUESTS, sc->ciss_cfg->max_outstanding_commands);
    
    if (1/*bootverbose*/)
	ciss_printf(sc, "using %d of %d available commands\n",
		    sc->ciss_max_requests, sc->ciss_cfg->max_outstanding_commands);

    /*
     * Create the DMA tag for commands.
     */
    if (bus_dma_tag_create(sc->ciss_parent_dmat,	/* parent */
			   1, 0, 			/* alignment, boundary */
			   BUS_SPACE_MAXADDR,		/* lowaddr */
			   BUS_SPACE_MAXADDR, 		/* highaddr */
			   NULL, NULL, 			/* filter, filterarg */
			   sizeof(struct ciss_command) * 
			   sc->ciss_max_requests, 1,	/* maxsize, nsegments */
			   BUS_SPACE_MAXSIZE_32BIT,	/* maxsegsize */
			   0,				/* flags */
			   &sc->ciss_command_dmat)) {
	ciss_printf(sc, "can't allocate command DMA tag\n");
	return(ENOMEM);
    }
    /*
     * Allocate memory and make it available for DMA.
     */
    if (bus_dmamem_alloc(sc->ciss_command_dmat, (void **)&sc->ciss_command, 
			 BUS_DMA_NOWAIT, &sc->ciss_command_map)) {
	ciss_printf(sc, "can't allocate command memory\n");
	return(ENOMEM);
    }
    bus_dmamap_load(sc->ciss_command_dmat, sc->ciss_command_map, sc->ciss_command, 
		    sizeof(struct ciss_command) * sc->ciss_max_requests, ciss_command_map_helper, sc, 0);
    bzero(sc->ciss_command, sizeof(*sc->ciss_command));

    /*/
     * Set up the request and command structures, push requests onto the free queue.
     */
    for (i = 0, cr = &sc->ciss_request[0]; i < sc->ciss_max_requests; i++, cr++) {
	cr->cr_sc = sc;
	cr->cr_tag = i;
	ciss_enqueue_free(cr);
    }
    return(0);
}

static void
ciss_command_map_helper(void *arg, bus_dma_segment_t *segs, int nseg, int error)
{
    struct ciss_softc	*sc = (struct ciss_softc *)arg;

    sc->ciss_command_phys = segs->ds_addr;
}

/************************************************************************
 * Find logical drives on the adapter.
 */
static int
ciss_init_logical(struct ciss_softc *sc)
{
    struct ciss_request		*cr;
    struct ciss_command		*cc;
    struct ciss_report_cdb	*crc;
    struct ciss_lun_report	*cll;
    int				error, i;
    size_t			report_size;
    int				ndrives;
    int				command_status;

    debug_called(2);

    cr = NULL;
    cll = NULL;

    /*
     * Get a reqest, allocate storage for the address list.
     */
    if ((error = ciss_get_request(sc, &cr)) != 0)
	goto out;
    cc = CISS_FIND_COMMAND(cr);
    crc = (struct ciss_report_cdb *)&(cc->cdb.cdb[0]);
    report_size = sizeof(*cll) + CISS_MAX_LOGICAL * sizeof(union ciss_device_address);
    if ((cll = malloc(report_size, CISS_MALLOC_CLASS, M_NOWAIT | M_ZERO)) == NULL) {
	ciss_printf(sc, "can't allocate memory for logical drive list\n");
	error = ENOMEM;
	goto out;
    }

    /*
     * Build the Report Logical LUNs command.
     */
    cr->cr_data = cll;
    cr->cr_length = report_size;
    cr->cr_flags = CISS_REQ_DATAIN;

    bzero(crc, sizeof(*crc));
    
    cc->header.address.physical.mode = CISS_HDR_ADDRESS_MODE_PERIPHERAL;
    cc->header.address.physical.bus = 0;
    cc->header.address.physical.target = 0;
    cc->cdb.cdb_length = sizeof(*crc);
    cc->cdb.type = CISS_CDB_TYPE_COMMAND;
    cc->cdb.attribute = CISS_CDB_ATTRIBUTE_SIMPLE;
    cc->cdb.direction = CISS_CDB_DIRECTION_READ;
    cc->cdb.timeout = 30;	/* XXX better suggestions? */
    crc->opcode = CISS_OPCODE_REPORT_LOGICAL_LUNS;
    crc->length = htonl(report_size);			/* big-endian field */
    cll->list_size = htonl(report_size - sizeof(*cll));	/* big-endian field */
    
    /*
     * Submit the request and busy-wait for it to complete.
     * (timeout here should be much greater than above)
     */
    if ((error = ciss_poll_request(cr, 60 * 1000)) != 0) {
	ciss_printf(sc, "error sending Report Logical LUNs command (%d)\n", error);
	goto out;
    }

    /*
     * Check response.  Note that data over/underrun is OK.
     */
    ciss_report_request(cr, &command_status, NULL);
    switch(command_status) {
    case CISS_CMD_STATUS_SUCCESS:		/* buffer right size */
    case CISS_CMD_STATUS_DATA_UNDERRUN:	/* buffer too large, not bad */
	break;
    case CISS_CMD_STATUS_DATA_OVERRUN:
	ciss_printf(sc, "WARNING: more logical drives than driver limit (%d), adjust CISS_MAX_LOGICAL\n",
		    CISS_MAX_LOGICAL);
	break;
    default:
	ciss_printf(sc, "error detecting logical drive configuration (%d)\n",
		    command_status);
	return(EIO);
    }

    /* sanity-check reply */
    ndrives = (ntohl(cll->list_size) / sizeof(union ciss_device_address));
    if ((ndrives < 0) || (ndrives > CISS_MAX_LOGICAL)) {
	ciss_printf(sc, "adapter claims to report absurd number of logical drives (%d > %d)\n",
		    ndrives, CISS_MAX_LOGICAL);
	return(ENXIO);
    }

    /*
     * Save logical drive information.
     */
    if (1/*bootverbose*/)
	ciss_printf(sc, "%d logical drive%s\n", ndrives, (ndrives > 1) ? "s" : "");
    for (i = 0; i < CISS_MAX_LOGICAL; i++) {
	if (i < ndrives) {
	    sc->ciss_logical[i].cl_address = cll->lun[i];	/* XXX endianness? */
	    sc->ciss_logical[i].cl_status = CISS_LD_ONLINE;
	    /* XXX get more information on drive here */
	    if (1/*bootverbose*/)
		ciss_printf(sc, "logical drive %d at %d\n", i, cll->lun[i].logical.lun);
	} else {
	    sc->ciss_logical[i].cl_status = CISS_LD_NONEXISTENT;
	}
    }
    error = 0;
    
 out:
    /*
     * Note that if the error is a timeout, we are taking a slight risk
     * here that the adapter will not respond at a later time, scribbling
     * over host memory.
     */
    if (cr != NULL)
	ciss_release_request(cr);
    if (cll != NULL)
	free(cll, CISS_MALLOC_CLASS);
    return(error);
}

/************************************************************************
 * Release adapter resources.
 */
static void
ciss_free(struct ciss_softc *sc)
{

    debug_called(2);

    /* release I/O resources */
    if (sc->ciss_regs_resource != NULL)
	bus_release_resource(sc->ciss_dev, SYS_RES_MEMORY, sc->ciss_regs_rid, sc->ciss_regs_resource);
    if (sc->ciss_cfg_resource != NULL)
	bus_release_resource(sc->ciss_dev, SYS_RES_MEMORY, sc->ciss_cfg_rid, sc->ciss_cfg_resource);
    if (sc->ciss_intr != NULL)
	bus_teardown_intr(sc->ciss_dev, sc->ciss_irq_resource, sc->ciss_intr);
    if (sc->ciss_irq_resource != NULL)
	bus_release_resource(sc->ciss_dev, SYS_RES_IRQ, sc->ciss_irq_rid, sc->ciss_irq_resource);

    /* destroy DMA tags */
    if (sc->ciss_parent_dmat)
	bus_dma_tag_destroy(sc->ciss_parent_dmat);
    if (sc->ciss_buffer_dmat)
	bus_dma_tag_destroy(sc->ciss_buffer_dmat);

    /* destroy command memory and DMA tag */
    if (sc->ciss_command != NULL) {
	bus_dmamap_unload(sc->ciss_command_dmat, sc->ciss_command_map);
	bus_dmamem_free(sc->ciss_command_dmat, sc->ciss_command, sc->ciss_command_map);
    }
    if (sc->ciss_buffer_dmat)
	bus_dma_tag_destroy(sc->ciss_command_dmat);

    /* disconnect from CAM */
    xpt_bus_deregister(cam_sim_path(sc->ciss_cam_sim));
    if (sc->ciss_cam_sim)
	cam_sim_free(sc->ciss_cam_sim, 0);
    if (sc->ciss_cam_devq)
	cam_simq_free(sc->ciss_cam_devq);
    /* XXX what about ciss_cam_path? */
}

/************************************************************************
 * Give a command to the adapter.
 *
 * Note that this uses the simple transport layer directly.  If we want
 * to add support for other layers, we'll need a switch of some sort.
 *
 * Note that the simple transport layer has no way of refusing a 
 * command; we only have as many request structures as the adapter
 * supports commands, so we don't have to check (this presumes that
 * the adapter can handle commands as fast as we throw them at it).
 */
static int
ciss_start(struct ciss_request *cr)
{
    int			error;

    debug(4, "post command %d at 0x%x", cr->cr_tag, CISS_FIND_COMMANDPHYS(cr));

    /*
     * Map the request's data.
     */
    if ((error = ciss_map_request(cr)))
	return(error);

#if 0
    ciss_print_request(cr);
#endif

    /*
     * Post the command to the adapter.
     */
    CISS_TL_SIMPLE_POST_CMD(cr->cr_sc, CISS_FIND_COMMANDPHYS(cr));

    return(0);
}

/************************************************************************
 * Fetch completed request(s) from the adapter, queue them for 
 * completion handling.
 *
 * Note that this uses the simple transport layer directly.  If we want
 * to add support for other layers, we'll need a switch of some sort.
 *
 * Note that the simple transport mechanism does not require any
 * reentrancy protection; the OPQ read is atomic.  If there is a chance
 * of a race with something else that might move the request off the busy
 * list, then we will have to lock against that (eg. timeouts, etc.)
 */
static void
ciss_done(struct ciss_softc *sc)
{
    struct ciss_request	*cr;
    struct ciss_command	*cc;
    u_int32_t		tag, index;
    int			complete;
    
    debug_called(6);

   /*
     * Loop quickly taking requests from the adapter and moving them
     * from the busy queue to the completed queue.
     */
    complete = 0;
    for (;;) {
	tag = CISS_TL_SIMPLE_FETCH_CMD(sc);
	if (tag == CISS_TL_SIMPLE_OPQ_EMPTY)
	    break;
	debug(4, "completed 0x%08x", tag);
	index = tag >> 2;
	if (index >= sc->ciss_max_requests) {
	    ciss_printf(sc, "completed nonbusy request %d\n", index);
	    continue;
	}
	cr = &(sc->ciss_request[index]);
	cc = CISS_FIND_COMMAND(cr);
	cc->header.host_tag = tag;	/* not updated by adapter */
	ciss_remove_busy(cr);
	ciss_enqueue_complete(cr);
	complete = 1;
    }
    
    /*
     * Invoke completion processing.  If we can defer this out of
     * interrupt context, that'd be good.
     */
    if (complete)
	ciss_complete(sc);
}

/************************************************************************
 * Take an interrupt from the adapter.
 */
static void
ciss_intr(void *arg)
{
    struct ciss_softc	*sc = (struct ciss_softc *)arg;

    /*
     * The only interrupt we recognise indicates that there are
     * entries in the outbound post queue.
     */
    ciss_done(sc);
}

/************************************************************************
 * Process completed requests.
 *
 * Requests can be completed in three fashions:
 *
 * - by invoking a callback function (cr_complete is non-null)
 * - by waking up a sleeper (cr_flags has CISS_REQ_SLEEP set)
 * - by clearing the CISS_REQ_POLL flag in interrupt/timeout context
 */
static void
ciss_complete(struct ciss_softc *sc)
{
    struct ciss_request	*cr;

    debug_called(4);

    /*
     * Loop taking requests off the completed queue and performing
     * completion processing on them.
     */
    for (;;) {
	if ((cr = ciss_dequeue_complete(sc)) == NULL)
	    break;
	ciss_unmap_request(cr);
	
	/*
	 * If the request has a callback, invoke it.
	 */
	if (cr->cr_complete != NULL) {
	    cr->cr_complete(cr);
	    continue;
	}
	
	/*
	 * If someone is sleeping on this request, wake them up.
	 */
	if (cr->cr_flags & CISS_REQ_SLEEP) {
	    cr->cr_flags &= ~CISS_REQ_SLEEP;
	    wakeup(cr);
	    continue;
	}

	/*
	 * If someone is polling this request for completion, signal.
	 */
	if (cr->cr_flags & CISS_REQ_POLL) {
	    cr->cr_flags &= ~CISS_REQ_POLL;
	    continue;
	}
	
	/*
	 * Give up and throw the request back on the free queue.
	 * This should never happen; resources will probably be lost.
	 */
	ciss_printf(sc, "WARNING: completed command with no submitter\n");
	ciss_enqueue_free(cr);
    }
}

/************************************************************************
 * Report on the completion status of a request, and pass back
 * SCSI and command status values.
 */
static int
ciss_report_request(struct ciss_request *cr, int *command_status, int *scsi_status)
{
    struct ciss_command		*cc;
    struct ciss_error_info	*ce;

    debug_called(4);

    cc = CISS_FIND_COMMAND(cr);
    ce = (struct ciss_error_info *)&(cc->sg[0]);

    /*
     * We don't consider data under/overrun an error for the
     * Report Logical/Physical LUNs commands.
     */
    if ((cc->header.host_tag & CISS_HDR_HOST_TAG_ERROR) &&
	((cc->cdb.cdb[0] == CISS_OPCODE_REPORT_LOGICAL_LUNS) ||
	 (cc->cdb.cdb[0] == CISS_OPCODE_REPORT_PHYSICAL_LUNS))) {
	cc->header.host_tag &= ~CISS_HDR_HOST_TAG_ERROR;
	debug(4, "ignoring irrelevant under/overrun error");
    }
    
    /*
     * Check the command's error bit, if clear, there's no status and
     * everything is OK.
     */
    if (!(cc->header.host_tag & CISS_HDR_HOST_TAG_ERROR)) {
	if (scsi_status != NULL)
	    *scsi_status = SCSI_STATUS_OK;
	if (command_status != NULL)
	    *command_status = CISS_CMD_STATUS_SUCCESS;
	return(0);
    } else {
	if (command_status != NULL)
	    *command_status = ce->command_status;
	if (scsi_status != NULL) {
	    if (ce->command_status == CISS_CMD_STATUS_TARGET_STATUS) {
		*scsi_status = ce->scsi_status;
	    } else {
		*scsi_status = -1;
	    }
	}
	if (1/*bootverbose*/)
	    ciss_printf(cr->cr_sc, "command status 0x%x  scsi status 0x%x\n",
			ce->command_status, ce->scsi_status);
    }
    return(1);
}

/************************************************************************
 * Issue a request and poll for completion.
 *
 * Timeout in milliseconds.
 */
static int
ciss_poll_request(struct ciss_request *cr, int timeout)
{
    int		error;
    
    debug_called(3);

    cr->cr_flags |= CISS_REQ_POLL;
    if ((error = ciss_start(cr)) != 0)
	return(error);

    do {
	ciss_done(cr->cr_sc);
	if (!(cr->cr_flags & CISS_REQ_POLL))
	    return(0);
	DELAY(1000);
    } while (timeout-- >= 0);
    return(EWOULDBLOCK);
}

/************************************************************************
 * Issue a request and sleep waiting for completion.
 *
 * Timeout in milliseconds.
 * Note that a spurious wakeup will reset the timeout.
 */
static int
ciss_wait_request(struct ciss_request *cr, int timeout)
{
    int		s, error;

    debug_called(3);

    cr->cr_flags |= CISS_REQ_SLEEP;
    if ((error = ciss_start(cr)) != 0)
	return(error);

    s = splcam();
    while (cr->cr_flags & CISS_REQ_SLEEP) {
	error = tsleep(cr, PCATCH, "cisswait", (timeout * hz) / 1000);
	/* 
	 * On wakeup or interruption due to restartable activity, go back
	 * and check to see if we're done.
	 */
	if ((error == 0) || (error == ERESTART)) {
	    error = 0;
	    continue;
	}
	/*
	 * Timeout, interrupted system call, etc.
	 */
	break;
    }
    splx(s);
    return(error);
}

/************************************************************************
 * Fetch and initialise a request
 */
static int
ciss_get_request(struct ciss_softc *sc, struct ciss_request **crp)
{
    struct ciss_request *cr;
    struct ciss_command	*cc;
    u_int32_t		cmdphys;

    debug_called(4);

    /*
     * Get a request and clean it up.
     */
    if ((cr = ciss_dequeue_free(sc)) == NULL)
	return(ENOMEM);
    cr->cr_data = NULL;
    cr->cr_flags = 0;
    cr->cr_complete = NULL;
    
    /*
     * Clean up the command structure.
     *
     * Note that we set up the error_info structure here, since the
     * length can be overwritten by any command.
     */
    cc = CISS_FIND_COMMAND(cr);
    cc->header.sg_in_list = 0;		/* kinda inefficient this way */
    cc->header.sg_total = 0;
    cc->header.host_tag = cr->cr_tag << 2;
    cmdphys = CISS_FIND_COMMANDPHYS(cr);
    cc->error_info.error_info_address = cmdphys + sizeof(struct ciss_command);
    cc->error_info.error_info_length = CISS_COMMAND_ALLOC_SIZE - sizeof(struct ciss_command);
    
    *crp = cr;
    return(0);
}

/************************************************************************
 * Release a request to the free list.
 */
static void
ciss_release_request(struct ciss_request *cr)
{
    struct ciss_softc	*sc;
    int			s;

    sc = cr->cr_sc;
    
    /* release the request to the free queue */
    ciss_requeue_free(cr);

    /*
     * If the SIM queue has been frozen due to resource starvation, unfreeze it
     * and run it again immediately.
     */
    s = splcam();
    if (sc->ciss_flags & CISS_FLAG_SIMQ_FROZEN)
	xpt_release_simq(sc->ciss_cam_sim, 1);
    splx(s);
}

/************************************************************************
 * Map a request into bus-visible space, initialise the scatter/gather
 * list.
 */
static int
ciss_map_request(struct ciss_request *cr)
{
    struct ciss_softc	*sc;
    
    sc = cr->cr_sc;

    /* check that mapping is necessary */
    if ((cr->cr_flags & CISS_REQ_MAPPED) || (cr->cr_data == NULL))
	return(0);
    
    bus_dmamap_load(sc->ciss_buffer_dmat, cr->cr_datamap, cr->cr_data, cr->cr_length,
		    ciss_request_map_helper, CISS_FIND_COMMAND(cr), 0);
	
    if (cr->cr_flags & CISS_REQ_DATAIN)
	bus_dmamap_sync(sc->ciss_buffer_dmat, cr->cr_datamap, BUS_DMASYNC_PREREAD);
    if (cr->cr_flags & CISS_REQ_DATAOUT)
	bus_dmamap_sync(sc->ciss_buffer_dmat, cr->cr_datamap, BUS_DMASYNC_PREWRITE);

    cr->cr_flags |= CISS_REQ_MAPPED;
    return(0);
}

static void
ciss_request_map_helper(void *arg, bus_dma_segment_t *segs, int nseg, int error)
{
    struct ciss_command	*cc;
    int			i;
    
    cc = (struct ciss_command *)arg;
    for (i = 0; i < nseg; i++) {
	cc->sg[i].address = segs[i].ds_addr;
	cc->sg[i].length = segs[i].ds_len;
	cc->sg[i].extension = 0;
    }
    /* we leave the s/g table entirely within the command */
    cc->header.sg_in_list = nseg;
    cc->header.sg_total = nseg;
}

/************************************************************************
 * Unmap a reqest from bus-visible space.
 */
static void
ciss_unmap_request(struct ciss_request *cr)
{
    struct ciss_softc	*sc;
    
    sc = cr->cr_sc;

    /* check that unmapping is necessary */
    if (!(cr->cr_flags & CISS_REQ_MAPPED) || (cr->cr_data == NULL))
	return;

    if (cr->cr_flags & CISS_REQ_DATAIN)
	bus_dmamap_sync(sc->ciss_buffer_dmat, cr->cr_datamap, BUS_DMASYNC_POSTREAD);
    if (cr->cr_flags & CISS_REQ_DATAOUT)
	bus_dmamap_sync(sc->ciss_buffer_dmat, cr->cr_datamap, BUS_DMASYNC_POSTWRITE);

    bus_dmamap_unload(sc->ciss_buffer_dmat, cr->cr_datamap);
}

/************************************************************************
 * Attach the driver to CAM.
 *
 * We put all the logical drives on a single SCSI bus.
 */
static int
ciss_cam_init(struct ciss_softc *sc)
{

    /*
     * Allocate a devq.  We can reuse this for the masked physical devices if we
     * decide to export these as well.
     */
    if ((sc->ciss_cam_devq = cam_simq_alloc(sc->ciss_max_requests)) == NULL) {
	ciss_printf(sc, "can't allocate CAM SIM queue\n");
	return(ENOMEM);
    }

    /*
     * Create a SIM.
     */
    if ((sc->ciss_cam_sim = cam_sim_alloc(ciss_cam_action, ciss_cam_poll, "ciss", sc,
				      device_get_unit(sc->ciss_dev), 1,
				      sc->ciss_cfg->max_outstanding_commands, sc->ciss_cam_devq)) == NULL) {
	ciss_printf(sc, "can't allocate CAM SIM\n");
	return(ENOMEM);
    }

    /*
     * Register bus 0 (the 'logical drives' bus) with this SIM.
     */
    if (xpt_bus_register(sc->ciss_cam_sim, 0) != 0) {
	ciss_printf(sc, "can't register SCSI bus 0\n");
	return(ENXIO);
    }

    /*
     * Initiate a rescan of the bus.
     */
    ciss_cam_rescan(sc);
    
    return(0);
}

/************************************************************************
 * Initiate a rescan of the 'logical devices' SIM
 */ 
static void
ciss_cam_rescan(struct ciss_softc *sc)
{
    union ccb	*ccb;

    if ((ccb = malloc(sizeof(union ccb), M_TEMP, M_WAITOK | M_ZERO)) == NULL) {
	ciss_printf(sc, "rescan failed (can't allocate CCB)\n");
	return;
    }
    
    if (xpt_create_path(&sc->ciss_cam_path, xpt_periph, cam_sim_path(sc->ciss_cam_sim), 0, 0)
	!= CAM_REQ_CMP) {
	ciss_printf(sc, "rescan failed (can't create path)\n");
	return;
    }
    
    xpt_setup_ccb(&ccb->ccb_h, sc->ciss_cam_path, 5/*priority (low)*/);
    ccb->ccb_h.func_code = XPT_SCAN_BUS;
    ccb->ccb_h.cbfcnp = ciss_cam_rescan_callback;
    ccb->crcn.flags = CAM_FLAG_NONE;
    xpt_action(ccb);
 
    /* scan is now in progress */
}

static void
ciss_cam_rescan_callback(struct cam_periph *periph, union ccb *ccb)
{
    free(ccb, M_TEMP);
}

/************************************************************************
 * Handle requests coming from CAM
 */
static void
ciss_cam_action(struct cam_sim *sim, union ccb *ccb)
{
    switch (ccb->ccb_h.func_code) {

	/* perform SCSI I/O */
    case XPT_SCSI_IO:
	if (!ciss_cam_action_io(sim, (struct ccb_scsiio *)&ccb->csio))
	    return;
	break;

	/* perform geometry calculations */
    case XPT_CALC_GEOMETRY:
    {
	struct ccb_calc_geometry	*ccg = &ccb->ccg;
        u_int32_t			secs_per_cylinder;

	debug(2, "XPT_CALC_GEOMETRY %d:%d:%d", cam_sim_bus(sim), ccb->ccb_h.target_id, ccb->ccb_h.target_lun);

	/*
	 * This is the default geometry; hopefully we will have successfully talked
	 * to the 'disk' and obtained its private settings.
	 */
	ccg->heads = 255;
	ccg->secs_per_track = 32;
	secs_per_cylinder = ccg->heads * ccg->secs_per_track;
        ccg->cylinders = ccg->volume_size / secs_per_cylinder;
        ccb->ccb_h.status = CAM_REQ_CMP;
        break;
    }

	/* handle path attribute inquiry */
    case XPT_PATH_INQ:
    {
	struct ccb_pathinq	*cpi = &ccb->cpi;

	debug(2, "XPT_PATH_INQ %d:%d:%d", cam_sim_bus(sim), ccb->ccb_h.target_id, ccb->ccb_h.target_lun);

	cpi->version_num = 1;
	cpi->hba_inquiry = PI_TAG_ABLE;	/* XXX is this correct? */
	cpi->target_sprt = 0;
	cpi->hba_misc = 0;
	cpi->max_target = CISS_MAX_LOGICAL;
	cpi->max_lun = 0;		/* 'logical drive' channel only */
	cpi->initiator_id = CISS_MAX_LOGICAL;
	strncpy(cpi->sim_vid, "FreeBSD", SIM_IDLEN);
        strncpy(cpi->hba_vid, "msmith@freebsd.org", HBA_IDLEN);
        strncpy(cpi->dev_name, cam_sim_name(sim), DEV_IDLEN);
        cpi->unit_number = cam_sim_unit(sim);
        cpi->bus_id = cam_sim_bus(sim);
	cpi->base_transfer_speed = 132 * 1024;	/* XXX what to set this to? */
	ccb->ccb_h.status = CAM_REQ_CMP;
	break;
    }

    case XPT_GET_TRAN_SETTINGS:
    {
	struct ccb_trans_settings	*cts = &ccb->cts;
	int				bus, target;

	bus = cam_sim_bus(sim);
	target = cts->ccb_h.target_id;

	debug(2, "XPT_GET_TRAN_SETTINGS %d:%d", bus, target);
	cts->valid = 0;

	/* disconnect always OK */
	cts->flags |= CCB_TRANS_DISC_ENB;
	cts->valid |= CCB_TRANS_DISC_VALID;

	cts->ccb_h.status = CAM_REQ_CMP;
	break;
    }

    default:		/* we can't do this */
	debug(2, "unspported func_code = 0x%x", ccb->ccb_h.func_code);
	ccb->ccb_h.status = CAM_REQ_INVALID;
	break;
    }

    xpt_done(ccb);
}

/************************************************************************
 * Handle a CAM SCSI I/O request.
 */
static int
ciss_cam_action_io(struct cam_sim *sim, struct ccb_scsiio *csio)
{
    struct ciss_softc	*sc;
    int			bus, target;
    struct ciss_request	*cr;
    struct ciss_command	*cc;
    int			error;

    sc = cam_sim_softc(sim);
    bus = cam_sim_bus(sim);
    target = csio->ccb_h.target_id;

    debug(2, "XPT_SCSI_IO %d:%d:%d", bus, target, csio->ccb_h.target_lun);

    /* check for I/O attempt to nonexistent device */
    if ((bus != 0) ||
	(target > CISS_MAX_LOGICAL) ||
	(sc->ciss_logical[target].cl_status == CISS_LD_NONEXISTENT)) {
	debug(2, "  device does not exist");
	csio->ccb_h.status = CAM_REQ_CMP_ERR;
    }

    /* firmware does not support commands > 10 bytes */
    if (csio->cdb_len > 12/*CISS_CDB_BUFFER_SIZE*/) {
	debug(2, "  command too large (%d > %d)", csio->cdb_len, CISS_CDB_BUFFER_SIZE);
	csio->ccb_h.status = CAM_REQ_CMP_ERR;
    }

    /* check that the CDB pointer is not to a physical address */
    if ((csio->ccb_h.flags & CAM_CDB_POINTER) && (csio->ccb_h.flags & CAM_CDB_PHYS)) {
	debug(2, "  CDB pointer is to physical address");
	csio->ccb_h.status = CAM_REQ_CMP_ERR;
    }

    /* if there is data transfer, it must be to/from a virtual address */
    if ((csio->ccb_h.flags & CAM_DIR_MASK) != CAM_DIR_NONE) {
	if (csio->ccb_h.flags & CAM_DATA_PHYS) {		/* we can't map it */
	    debug(2, "  data pointer is to physical address");
	    csio->ccb_h.status = CAM_REQ_CMP_ERR;
	}
	if (csio->ccb_h.flags & CAM_SCATTER_VALID) {	/* we want to do the s/g setup */
	    debug(2, "  data has premature s/g setup");
	    csio->ccb_h.status = CAM_REQ_CMP_ERR;
	}
    }

    /* abandon aborted ccbs or those that have failed validation */
    if ((csio->ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_INPROG) {
	debug(2, "abandoning CCB due to abort/validation failure");
	return(EINVAL);
    }

    /*
     * Get a request to manage this command.  If we can't, return the ccb,
     * freeze the queue and flag so that we unfreeze it when a request completes.
     */
    if ((error = ciss_get_request(sc, &cr)) != 0) {
	ciss_cam_backoff_ccb(sc, csio);
	return(error);
    }
    cc = CISS_FIND_COMMAND(cr);

    /*
     * Build the command.
     */
    cr->cr_data = csio->data_ptr;
    cr->cr_length = csio->dxfer_len;
    cr->cr_complete = ciss_cam_complete;
    cr->cr_private = csio;
	
    cc->header.address.logical.mode = CISS_HDR_ADDRESS_MODE_LOGICAL;
    cc->header.address.logical.lun = target;
    cc->cdb.cdb_length = csio->cdb_len;
    cc->cdb.type = CISS_CDB_TYPE_COMMAND;
    cc->cdb.attribute = CISS_CDB_ATTRIBUTE_SIMPLE;	/* XXX ordered tags? */
    if ((csio->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_OUT) {
	cr->cr_flags = CISS_REQ_DATAOUT;
	cc->cdb.direction = CISS_CDB_DIRECTION_WRITE;
    } else if ((csio->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_IN) {
	cr->cr_flags = CISS_REQ_DATAIN;
	cc->cdb.direction = CISS_CDB_DIRECTION_READ;
    } else {
	cr->cr_flags = 0;
	cc->cdb.direction = CISS_CDB_DIRECTION_NONE;
    }
    cc->cdb.timeout = (csio->ccb_h.timeout / 1000) + 1;
    if (csio->ccb_h.flags & CAM_CDB_POINTER) {
	bcopy(csio->cdb_io.cdb_ptr, &cc->cdb.cdb[0], csio->cdb_len);
    } else {
	bcopy(csio->cdb_io.cdb_bytes, &cc->cdb.cdb[0], csio->cdb_len);
    }

    /*
     * Submit the request to the adapter.
     *
     * Note that this may fail if we're unable to map the request (and if
     * we ever learn a transport layer other than simple, may fail if the
     * adapter rejects the command).
     */
    if ((error = ciss_start(cr)) != 0) {
	ciss_cam_backoff_ccb(sc, csio);
	ciss_release_request(cr);
	return(error);
    }
	
    return(0);
}

/************************************************************************
 * Return a ccb due to resource starvation.
 *
 * Note that the caller must eventually call xpt_done on the ccb.
 */
static void
ciss_cam_backoff_ccb(struct ciss_softc *sc, struct ccb_scsiio *csio)
{
    int		s;
    
    s = splcam();
    sc->ciss_flags |= CISS_FLAG_SIMQ_FROZEN;
    xpt_freeze_simq(sc->ciss_cam_sim, 1);
    splx(s);
    csio->ccb_h.status |= CAM_REQUEUE_REQ;
}


/************************************************************************
 * Check for possibly-completed commands.
 */
static void
ciss_cam_poll(struct cam_sim *sim)
{
    struct ciss_softc	*sc = cam_sim_softc(sim);

    debug_called(2);

    ciss_done(sc);
}

/************************************************************************
 * Handle completion of a command - pass results back through the CCB
 */
static void
ciss_cam_complete(struct ciss_request *cr)
{
    struct ciss_softc		*sc;
    struct ciss_command		*cc;
    struct ciss_error_info	*ce;
    struct ccb_scsiio		*csio;
    int				scsi_status;
    int				command_status;

    debug_called(2);

    sc = cr->cr_sc;
    cc = CISS_FIND_COMMAND(cr);
    ce = (struct ciss_error_info *)&(cc->sg[0]);
    csio = (struct ccb_scsiio *)cr->cr_private;

    /*
     * Extract status values from request.
     */
    ciss_report_request(cr, &command_status, &scsi_status);
    csio->scsi_status = scsi_status;
    
    /*
     * Handle specific SCSI status values.
     */
    switch(scsi_status) {
	/* no status due to controller error */
    case -1:				
	debug(2, "controller error");
	csio->ccb_h.status = CAM_REQ_CMP_ERR;
	break;
	
	/* no status due to command completed OK */
    case SCSI_STATUS_OK:		/* CISS_SCSI_STATUS_GOOD */
	debug(2, "SCSI_STATUS_OK");
	csio->ccb_h.status = CAM_REQ_CMP;
	break;

	/* check condition, sense data included */
    case SCSI_STATUS_CHECK_COND:	/* CISS_SCSI_STATUS_CHECK_CONDITION */
	debug(2, "SCSI_STATUS_CHECK_COND  sense %d  resid %d", ce->sense_length, ce->residual_count);
	bzero(&csio->sense_data, SSD_FULL_SIZE);
	bcopy(&ce->sense_info[0], &csio->sense_data, ce->sense_length);
	csio->sense_len = ce->sense_length;
	csio->resid = ce->residual_count;
	csio->ccb_h.status = CAM_SCSI_STATUS_ERROR | CAM_AUTOSNS_VALID;
	break;

    case SCSI_STATUS_BUSY:		/* CISS_SCSI_STATUS_BUSY */
	debug(2, "SCSI_STATUS_BUSY");
	csio->ccb_h.status = CAM_SCSI_BUSY;
	break;

    default:
	debug(2, "unknown status 0x%x", csio->scsi_status);
	csio->ccb_h.status = CAM_REQ_CMP_ERR;
	break;
    }
    xpt_done((union ccb *)csio);
    ciss_release_request(cr);
}

/************************************************************************
 * Periodic status monitoring.
 */
static void
ciss_periodic(void *arg)
{
    struct ciss_softc	*sc;
    int			s;
    
    sc = (struct ciss_softc *)arg;

    /*
     * Check the controller heartbeat.
     */
    if (sc->ciss_cfg->heartbeat == sc->ciss_heartbeat) {
	sc->ciss_heart_attack++;
	if (sc->ciss_heart_attack == 3) {
	    ciss_printf(sc, "CONTROLLER HEARTBEAT FAILED\n");
	    /* XXX should reset controller here */
	}
    } else {
	sc->ciss_heartbeat = sc->ciss_cfg->heartbeat;
	sc->ciss_heart_attack = 0;
    }
    
    /*
     * If the SIM queue has been frozen due to resource starvation, unfreeze it
     * and run it again immediately.  This check serves primarily to ensure that
     * there's no possibility of a race between freezing the queue and unfreezing
     * it in ciss_release_request.
     */
    s = splcam();
    if (sc->ciss_flags & CISS_FLAG_SIMQ_FROZEN)
	xpt_release_simq(sc->ciss_cam_sim, 1);
    splx(s);

    /*
     * If the notify event request has died for some reason, or has not
     * started yet, restart it.
     */
    if (~(sc->ciss_flags & CISS_FLAG_NOTIFY_OK)) {
	ciss_notify_event(sc, NULL);
    }
}

/************************************************************************
 * Request a notification response from the controller.
 *
 * If (cr) is NULL, this is the first request of the controller, so
 * reset the controller's message pointer and start with the oldest
 * message available.
 */
static void
ciss_notify_event(struct ciss_softc *sc, struct ciss_request *cr)
{
    struct ciss_command		*cc;
    struct ciss_notify_cdb	*cnc;
    int				error, seek_to_oldest;
    
    /* get a request if we don't already have one, and set it up */
    if (cr == NULL) {
	if ((error = ciss_get_request(sc, &cr)) != 0) {
	    sc->ciss_flags &= ~CISS_FLAG_NOTIFY_OK;
	    ciss_printf(sc, "can't get notify event request\n");
	    return;
	}
	/* get a databuffer */
	if ((cr->cr_data = malloc(CISS_NOTIFY_DATA_SIZE, CISS_MALLOC_CLASS, M_NOWAIT)) == NULL) {
	    sc->ciss_flags &= ~CISS_FLAG_NOTIFY_OK;
	    ciss_release_request(cr);
	    ciss_printf(sc, "can't get notify event request buffer\n");
	    return;
	}
	cr->cr_length = CISS_NOTIFY_DATA_SIZE;
	seek_to_oldest = 1;
    } else {
	seek_to_oldest = 0;
    }
    
    cc = CISS_FIND_COMMAND(cr);
    cnc = (struct ciss_notify_cdb *)&(cc->cdb.cdb[0]);
    bzero(cr->cr_data, CISS_NOTIFY_DATA_SIZE);
    
    /* re-setup the request (since we haven't released it) */
    cr->cr_flags = 0;
    cr->cr_complete = ciss_notify_complete;

    /* (re)build the notify event command */
    cc->header.address.physical.mode = CISS_HDR_ADDRESS_MODE_PERIPHERAL;
    cc->header.address.physical.bus = 0;
    cc->header.address.physical.target = 0;

    cc->cdb.cdb_length = sizeof(*cnc);
    cc->cdb.type = CISS_CDB_TYPE_COMMAND;
    cc->cdb.attribute = CISS_CDB_ATTRIBUTE_SIMPLE;
    cc->cdb.direction = CISS_CDB_DIRECTION_READ;
    cc->cdb.timeout = 0;	/* no timeout, we hope */
    
    cnc->opcode = CISS_OPCODE_READ;
    cnc->command = CISS_COMMAND_NOTIFY_ON_EVENT;
    cnc->synchronous = 0;
    cnc->ordered = 1;
    cnc->seek_to_oldest = seek_to_oldest;
    cnc->new_only = 0;
    cnc->timeout = 0;		/* no timeout, we hope */
    cnc->length = CISS_NOTIFY_DATA_SIZE;

    /* submit the request */
    if ((error = ciss_start(cr)) != 0) {
	sc->ciss_flags &= ~CISS_FLAG_NOTIFY_OK;
	free(cr->cr_data, CISS_MALLOC_CLASS);
	ciss_release_request(cr);
	ciss_printf(sc, "can't submit notify event request\n");
    }
    sc->ciss_flags |= CISS_FLAG_NOTIFY_OK;
}

/************************************************************************
 * Handle a response to the Notify Event command.
 */
#if 0
struct 
{
    int		class;
    int		subclass;
    int		detail;
    void	(* action)(struct ciss_softc *sc, struct ciss_notify *cn);
    char	*text;
} ciss_notify_action[] = {
    {0,  0,  0, NULL, NULL},		/* no event */
    {0,  0,  1, NULL, NULL},		/* notifier disabled */
    {0,  0,  2, NULL, NULL},		/* notifier disabled */
    {0,  1,  0, ciss_notify_action_print, "event log overflow"},
    {1,  0,  0, ciss_notify_action_addressable, "physical drive removed"},
    {1,  0,  1, ciss_notify_action_addressable, "physical drive inserted"},
    {1,  1,  0, ciss_notify_action_addressable, "power supply removed"},
    {1,  1,  1, ciss_notify_action_addressable, "power supply inserted"},
    {1,  2,  0, ciss_notify_action_addressable, "fan removed"},
    {1,  2,  1, ciss_notify_action_addressable, "fan inserted"},
    {1,  3,  0, ciss_notify_action_addressable, "AC power lost"},	/* XXX box id irrelevant */
    {1,  3,  1, ciss_notify_action_addressable, "AC power supplied"},	/* XXX box id irrelevant */
    {1,  4, -4, ciss_notify_action_redundant_change, NULL},
    {2,  0, -1, ciss_notify_action_failure, "cable"},
    {2,  1, -1, ciss_notify_action_failure, "memory"},
    {2,  2, -1, ciss_notify_action_fan_failure, NULL},
    {2,  3, -1, ciss_notify_action_vrm_event, NULL},
    {3,  0, -1, ciss_notify_action_temperature, NULL},
    {3,  1, -1, ciss_notify_action_power_supply, NULL},
    {3,  2, -1, NULL, ciss_notify_action_   },
    {0,  0,  0, NULL, ciss_notify_action_   },
    {0,  0,  0, NULL, ciss_notify_action_   },
    {0,  0,  0, NULL, ciss_notify_action_   },
    {0,  0,  0, NULL, ciss_notify_action_   },
    {0,  0,  0, NULL, ciss_notify_action_   },
    
    {0, 0, 0, NULL, NULL}
};
#endif
    
static void
ciss_notify_complete(struct ciss_request *cr)
{
    struct ciss_command	*cc;
    struct ciss_notify	*cn;
    struct ciss_softc	*sc;
    int			scsi_status;
    int			command_status;
    
    cc = CISS_FIND_COMMAND(cr);
    cn = (struct ciss_notify *)cr->cr_data;
    sc = cr->cr_sc;
    
    /*
     * Report request results, decode status.
     */
    ciss_report_request(cr, &command_status, &scsi_status);

    /*
     * Abort the chain on a fatal error.
     */
    if ((command_status != CISS_CMD_STATUS_SUCCESS) && (command_status != CISS_CMD_STATUS_TIMEOUT)) {
	sc->ciss_flags &= ~CISS_FLAG_NOTIFY_OK;
	free(cr->cr_data, CISS_MALLOC_CLASS);
	ciss_release_request(cr);	
	ciss_printf(sc, "fatal error in notify event command\n");
	return;
    }
    
    {
	/* XXX print event */
    }
    
    /*
     * Send a new notify event command.
     */
    ciss_notify_event(sc, cr);
}


/************************************************************************
 * Print adapter information
 */
static void
ciss_print_adapter(struct ciss_softc *sc)
{
    ciss_printf(sc, "  signature '%.4s'\n", sc->ciss_cfg->signature);
    ciss_printf(sc, "  valence %d\n", sc->ciss_cfg->valence);
    ciss_printf(sc, "  supported I/O methods 0x%b\n", sc->ciss_cfg->supported_methods, "\20\2simple\n");
    ciss_printf(sc, "  active I/O method 0x%b\n", sc->ciss_cfg->active_method, "\20\2simple\n");
    ciss_printf(sc, "  4G page base 0x%08x\n", sc->ciss_cfg->command_physlimit);
    ciss_printf(sc, "  interrupt coalesce delay %dus\n", sc->ciss_cfg->interrupt_coalesce_delay);
    ciss_printf(sc, "  interrupt coalesce count %d\n", sc->ciss_cfg->interrupt_coalesce_count);
    ciss_printf(sc, "  max outstanding commands %d\n", sc->ciss_cfg->max_outstanding_commands);
    ciss_printf(sc, "  bus types 0x%b\n", sc->ciss_cfg->bus_types, 
		"\20\1ultra2\2ultra3\10fibre1\11fibre2\n");
    ciss_printf(sc, "  server name '%.16s'\n", sc->ciss_cfg->server_name);
    ciss_printf(sc, "  heartbeat 0x%x\n", sc->ciss_cfg->heartbeat);
}

/************************************************************************
 * Print a request.
 */
static void
ciss_print_request(struct ciss_request *cr)
{
    struct ciss_softc	*sc;
    struct ciss_command	*cc;
    int			i;

    sc = cr->cr_sc;
    cc = CISS_FIND_COMMAND(cr);
    
    ciss_printf(sc, "REQUEST @ %p\n", cr);
    ciss_printf(sc, "  data %p/%d  tag %d  flags %b\n",
	      cr->cr_data, cr->cr_length, cr->cr_tag, cr->cr_flags,
	      "\20\1mapped\2sleep\3poll\4dataout\5datain\n");
    ciss_printf(sc, "  sg list/total %d/%d  host tag 0x%x\n",
		cc->header.sg_in_list, cc->header.sg_total, cc->header.host_tag);
    switch(cc->header.address.mode.mode) {
    case CISS_HDR_ADDRESS_MODE_PERIPHERAL:
    case CISS_HDR_ADDRESS_MODE_MASK_PERIPHERAL:
	ciss_printf(sc, "  physical bus %d target %d\n",
		    cc->header.address.physical.bus, cc->header.address.physical.target);
	break;
    case CISS_HDR_ADDRESS_MODE_LOGICAL:
	ciss_printf(sc, "  logical unit %d\n", cc->header.address.logical.lun);
	break;
    }
    ciss_printf(sc, "  %s cdb length %d type %s attribute %s\n", 
		(cc->cdb.direction == CISS_CDB_DIRECTION_NONE) ? "no-I/O" :
		(cc->cdb.direction == CISS_CDB_DIRECTION_READ) ? "READ" :
		(cc->cdb.direction == CISS_CDB_DIRECTION_WRITE) ? "WRITE" : "??",
		cc->cdb.cdb_length,
		(cc->cdb.type == CISS_CDB_TYPE_COMMAND) ? "command" :
		(cc->cdb.type == CISS_CDB_TYPE_MESSAGE) ? "message" : "??",
		(cc->cdb.attribute == CISS_CDB_ATTRIBUTE_UNTAGGED) ? "untagged" :
		(cc->cdb.attribute == CISS_CDB_ATTRIBUTE_SIMPLE) ? "simple" :
		(cc->cdb.attribute == CISS_CDB_ATTRIBUTE_HEAD_OF_QUEUE) ? "head-of-queue" :
		(cc->cdb.attribute == CISS_CDB_ATTRIBUTE_ORDERED) ? "ordered" :
		(cc->cdb.attribute == CISS_CDB_ATTRIBUTE_AUTO_CONTINGENT) ? "auto-contingent" : "??");
    ciss_printf(sc, "  %*D\n", cc->cdb.cdb_length, &cc->cdb.cdb[0], " ");

    if (cc->header.host_tag & CISS_HDR_HOST_TAG_ERROR) {
	/* XXX print error info */
    } else {
	/* since we don't use chained s/g, don't support it here */
	for (i = 0; i < cc->header.sg_in_list; i++) {
	    if ((i % 4) == 0)
		ciss_printf(sc, "   ");
	    printf("0x%08x/%d ", (u_int32_t)cc->sg[i].address, cc->sg[i].length);
	    if ((((i + 1) % 4) == 0) || (i == (cc->header.sg_in_list - 1)))
		printf("\n");
	}
    }
}

/************************************************************************
 * Handle an open on the control device.
 */
static int
ciss_open(dev_t dev, int flags, int fmt, struct proc *p)
{
    struct ciss_softc	*sc;
    
    sc = (struct ciss_softc *)dev->si_drv1;

    /* we might want to veto if someone already has us open */
	
    sc->ciss_flags |= CISS_CONTROL_OPEN;
    return(0);
}

/************************************************************************
 * Handle the last close on the control device.
 */
static int
ciss_close(dev_t dev, int flags, int fmt, struct proc *p)
{
    struct ciss_softc	*sc;

    sc = (struct ciss_softc *)dev->si_drv1;
    
    sc->ciss_flags &= ~CISS_CONTROL_OPEN;
    return (0);
}

/********************************************************************************
 * Handle controller-specific control operations.
 */
static int
ciss_ioctl(dev_t dev, u_long cmd, caddr_t addr, int32_t flag, struct proc *p)
{
    struct ciss_softc	*sc;

    sc = (struct ciss_softc *)dev->si_drv1;

    return(ENOTTY);
}
