/*-
 * Copyright (c) 2001 Michael Smith
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$FreeBSD$
 */

/*
 * CISS adapter driver datastructures
 */

/************************************************************************
 * Tunable parameters
 */

/*
 * There is no guaranteed upper bound on the number of concurrent commands
 * an adapter may claim to support.  Cap it at a reasonable value.
 */
#define CISS_MAX_REQUESTS	256

/*
 * Maximum number of logical drives we support.
 */
#define CISS_MAX_LOGICAL	15

/*
 * Interrupt reduction can be controlled by tuning the interrupt coalesce
 * delay and count paramters.  The delay (in microseconds) defers delivery
 * of interrupts to increase the chance of there being more than one
 * completed command ready when the interrupt is delivered.  The count
 * expedites the delivery of the interrupt when the given number of 
 * commands are ready.
 *
 * If the delay is set to 0, interrupts are delivered immediately.
 */
#define CISS_INTERRUPT_COALESCE_DELAY	1000
#define CISS_INTERRUPT_COALESCE_COUNT	16

/************************************************************************
 * Command queue statistics
 */

#define CISSQ_FREE	0
#define CISSQ_BUSY	1
#define CISSQ_COMPLETE	2
#define CISSQ_COUNT	3

struct ciss_qstat 
{
    u_int32_t	q_length;
    u_int32_t	q_max;
};


/************************************************************************
 * Driver data structures
 */

/*
 * Each command issued to the adapter is managed by a request structure.
 */
struct ciss_request
{
    TAILQ_ENTRY(ciss_request)	cr_link;

    struct ciss_softc		*cr_sc;		/* controller softc */
    void			*cr_data;	/* data buffer */
    u_int32_t			cr_length;	/* data length */
    bus_dmamap_t		cr_datamap;	/* DMA map for data */
    int				cr_tag;
    int				cr_flags;
#define CISS_REQ_MAPPED		(1<<0)		/* data mapped */
#define CISS_REQ_SLEEP		(1<<1)		/* submitter sleeping */
#define CISS_REQ_POLL		(1<<2)		/* submitter polling */
#define CISS_REQ_DATAOUT	(1<<3)		/* data host->adapter */
#define CISS_REQ_DATAIN		(1<<4)		/* data adapter->host */
    
    void			(* cr_complete)(struct ciss_request *);
    void			*cr_private;
};

/*
 * The adapter command structure is defined with a zero-length scatter/gather list
 * size.  In practise, we want space for a scatter-gather list, and we also want
 * to avoid having commands cross page boundaries. 
 *
 * Note that 512 bytes yields 28 scatter/gather entries, or the ability to map
 * (26 * PAGE_SIZE) + 2 bytes of data.  On x86, this is 104kB.  256 bytes would only 
 * yield 12 entries, giving a mere 40kB, too small.
 */

#define CISS_COMMAND_ALLOC_SIZE		512	/* XXX tune to get sensible s/g list length */
#define CISS_COMMAND_SG_LENGTH	((CISS_COMMAND_ALLOC_SIZE - sizeof(struct ciss_command)) \
				 / sizeof(struct ciss_sg_entry))

/*
 * Per-logical-drive data.
 */
struct ciss_ldrive 
{
    union ciss_device_address	cl_address;
    
    int				cl_status;
#define CISS_LD_NONEXISTENT	0
#define CISS_LD_ONLINE		1
#define CISS_LD_OFFLINE		2
};

/*
 * Per-adapter data
 */
struct ciss_softc 
{
    /* bus connections */
    device_t			ciss_dev;		/* bus attachment */
    dev_t			ciss_dev_t;		/* control device */

    struct resource		*ciss_regs_resource;	/* register interface window */
    int				ciss_regs_rid;		/* resource ID */
    bus_space_handle_t		ciss_regs_bhandle;	/* bus space handle */
    bus_space_tag_t		ciss_regs_btag;		/* bus space tag */

    struct resource		*ciss_cfg_resource;	/* config struct interface window */
    int				ciss_cfg_rid;		/* resource ID */
    struct ciss_config_table	*ciss_cfg;		/* config table in adapter memory */
    u_int32_t			ciss_heartbeat;		/* last heartbeat value */
    int				ciss_heart_attack;	/* number of times we have seen this value */

    struct resource		*ciss_irq_resource;	/* interrupt */
    int				ciss_irq_rid;		/* resource ID */
    void			*ciss_intr;		/* interrupt handle */

    bus_dma_tag_t		ciss_parent_dmat;	/* parent DMA tag */
    bus_dma_tag_t		ciss_buffer_dmat;	/* data buffer/command DMA tag */

    u_int32_t			ciss_interrupt_mask;	/* controller interrupt mask bits */

    int				ciss_max_requests;
    struct ciss_request		ciss_request[CISS_MAX_REQUESTS];	/* requests */
    struct ciss_command		*ciss_command;		/* command structures */
    bus_dma_tag_t		ciss_command_dmat;	/* command DMA tag */
    bus_dmamap_t		ciss_command_map;	/* command DMA map */
    u_int32_t			ciss_command_phys;	/* command array base address */
    TAILQ_HEAD(,ciss_request)	ciss_free;		/* requests available for reuse */
    TAILQ_HEAD(,ciss_request)	ciss_busy;		/* requests in the adapter */
    TAILQ_HEAD(,ciss_request)	ciss_complete;		/* requests which have been returned by the adapter */

    struct callout_handle	ciss_periodic;		/* periodic event handling */

    struct ciss_ldrive		ciss_logical[CISS_MAX_LOGICAL];

    struct cam_devq		*ciss_cam_devq;
    struct cam_sim		*ciss_cam_sim;
    struct cam_path		*ciss_cam_path;

    int				ciss_flags;
#define CISS_FLAG_SIMQ_FROZEN	(1<<0)		/* SIM queue frozen due to request starvation */
#define CISS_FLAG_NOTIFY_OK	(1<<1)		/* notify command running OK */
#define CISS_CONTROL_OPEN	(1<<2)		/* control device is open */

    struct ciss_qstat		ciss_qstat[CISSQ_COUNT];	/* queue statistics */
};

/*
 * Given a request tag, find the corresponding command in virtual or
 * physical space.
 */
#define CISS_FIND_COMMAND(cr)		(&(cr)->cr_sc->ciss_command[(cr)->cr_tag])
#define CISS_FIND_COMMANDPHYS(cr)	((cr)->cr_sc->ciss_command_phys + \
					 ((cr)->cr_tag * CISS_COMMAND_ALLOC_SIZE))

/************************************************************************
 * Debugging/diagnostic output.
 */

/*
 * Debugging levels:
 *  0 - quiet, only emit warnings
 *  1 - noisy, emit major function points and things done
 *  2 - extremely noisy, emit trace items in loops, etc.
 */
#ifdef CISS_DEBUG
# define debug(level, fmt, args...)							\
	do {										\
	    if (level <= CISS_DEBUG) printf("%s: " fmt "\n", __FUNCTION__ , ##args);	\
	} while(0)
# define debug_called(level)						\
	do {								\
	    if (level <= CISS_DEBUG) printf(__FUNCTION__ ": called\n");	\
	} while(0)
# define debug_struct(s)		printf("  SIZE %s: %d\n", #s, sizeof(struct s))
# define debug_union(s)			printf("  SIZE %s: %d\n", #s, sizeof(union s))
# define debug_field(s, f)		printf("  OFFSET %s.%s: %d\n", #s, #f, ((int)&(((struct s *)0)->f)))
# define debug_const(c)			printf("  CONST %s %d/0x%x\n", #c, c, c);
#else
# define debug(level, fmt, args...)
# define debug_called(level)
# define debug_struct(s)
#endif

#define ciss_printf(sc, fmt, args...)	device_printf(sc->ciss_dev, fmt , ##args)

/************************************************************************
 * Queue primitives
 */

#define CISSQ_ADD(sc, qname)					\
	do {							\
	    struct ciss_qstat *qs = &(sc)->ciss_qstat[qname];	\
								\
	    qs->q_length++;					\
	    if (qs->q_length > qs->q_max)			\
		qs->q_max = qs->q_length;			\
	} while(0)

#define CISSQ_REMOVE(sc, qname)    (sc)->ciss_qstat[qname].q_length--
#define CISSQ_INIT(sc, qname)			\
	do {					\
	    sc->ciss_qstat[qname].q_length = 0;	\
	    sc->ciss_qstat[qname].q_max = 0;	\
	} while(0)


#define CISSQ_REQUEST_QUEUE(name, index)				\
static __inline void							\
ciss_initq_ ## name (struct ciss_softc *sc)				\
{									\
    TAILQ_INIT(&sc->ciss_ ## name);					\
    CISSQ_INIT(sc, index);						\
}									\
static __inline void							\
ciss_enqueue_ ## name (struct ciss_request *cr)				\
{									\
    int		s;							\
									\
    s = splcam();							\
    TAILQ_INSERT_TAIL(&cr->cr_sc->ciss_ ## name, cr, cr_link);		\
    CISSQ_ADD(cr->cr_sc, index);					\
    splx(s);								\
}									\
static __inline void							\
ciss_requeue_ ## name (struct ciss_request *cr)				\
{									\
    int		s;							\
									\
    s = splcam();							\
    TAILQ_INSERT_HEAD(&cr->cr_sc->ciss_ ## name, cr, cr_link);		\
    CISSQ_ADD(cr->cr_sc, index);					\
    splx(s);								\
}									\
static __inline struct ciss_request *					\
ciss_dequeue_ ## name (struct ciss_softc *sc)				\
{									\
    struct ciss_request	*cr;						\
    int			s;						\
									\
    s = splcam();							\
    if ((cr = TAILQ_FIRST(&sc->ciss_ ## name)) != NULL) {		\
	TAILQ_REMOVE(&sc->ciss_ ## name, cr, cr_link);			\
	CISSQ_REMOVE(sc, index);					\
    }									\
    splx(s);								\
    return(cr);								\
}									\
static __inline void							\
ciss_remove_ ## name (struct ciss_request *cr)				\
{									\
    int			s;						\
									\
    s = splcam();							\
    TAILQ_REMOVE(&cr->cr_sc->ciss_ ## name, cr, cr_link);		\
    CISSQ_REMOVE(cr->cr_sc, index);					\
    splx(s);								\
}									\
struct hack

CISSQ_REQUEST_QUEUE(free, CISSQ_FREE);
CISSQ_REQUEST_QUEUE(busy, CISSQ_BUSY);
CISSQ_REQUEST_QUEUE(complete, CISSQ_COMPLETE);

